//
//  CSDatabaseCache.m
//  C64SceneBrowser
//
//  Created by Andreas Varga on 27.03.12.
//  Copyright (c) 2012 __MyCompanyName__. All rights reserved.
//

#import "CSDatabaseCache.h"
#import "CSDatabaseItem.h"
#import "CSReleaseItem.h"
#import "CSScenerItem.h"
#import "CSGroupItem.h"
#import "CSEventItem.h"


@implementation CSDatabaseCache

@synthesize releaseDictionary = _releaseDictionary;
@synthesize userDictionary = _userDictionary;
@synthesize scenerDictionary = _scenerDictionary;
@synthesize groupDictionary = _groupDictionary;
@synthesize eventDictionary = _eventDictionary;


// ----------------------------------------------------------------------------
+ (CSDatabaseCache*) sharedInstance
// ----------------------------------------------------------------------------
{
    static CSDatabaseCache* instance = nil;
    
    if (instance == nil)
        instance = [[CSDatabaseCache alloc] init];
    
    return instance;
}


// ----------------------------------------------------------------------------
- (id)init
// ----------------------------------------------------------------------------
{
    self = [super init];
    if (self)
    {
        _releaseDictionary = [NSMutableDictionary dictionary];
        _userDictionary = [NSMutableDictionary dictionary];
        _scenerDictionary = [NSMutableDictionary dictionary];
        _groupDictionary = [NSMutableDictionary dictionary]; 
        _eventDictionary = [NSMutableDictionary dictionary]; 
        
        NSString* documentsPath = [NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES) objectAtIndex:0];
        _releasesPath = [documentsPath stringByAppendingPathComponent:@"Releases"];
        _scenersPath = [documentsPath stringByAppendingPathComponent:@"Sceners"];
        _favoritesPath = [documentsPath stringByAppendingPathComponent:@"Favorites"];
     
        [self readFavorites];
        
        BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:_releasesPath isDirectory:NULL];
        if (!exists)
            [[NSFileManager defaultManager] createDirectoryAtPath:_releasesPath withIntermediateDirectories:YES attributes:nil error:NULL];

        exists = [[NSFileManager defaultManager] fileExistsAtPath:_scenersPath isDirectory:NULL];
        if (!exists)
            [[NSFileManager defaultManager] createDirectoryAtPath:_scenersPath withIntermediateDirectories:YES attributes:nil error:NULL];
    }
    return self;
}


// ----------------------------------------------------------------------------
- (NSString*) releasesPath
// ----------------------------------------------------------------------------
{
    return _releasesPath;
}


// ----------------------------------------------------------------------------
- (NSString*) scenersPath
// ----------------------------------------------------------------------------
{
    return _scenersPath;
}


// ----------------------------------------------------------------------------
- (CSDatabaseItem*) itemForType:(CSDatabaseItemType)type withIdentifier:(NSString*)identifier
// ----------------------------------------------------------------------------
{
    NSMutableDictionary* dictionary = nil;
    switch (type)
    {
        case CSDB_RELEASE:
            dictionary = _releaseDictionary;
            break;
            
        case CSDB_SCENER:
            dictionary = _scenerDictionary;
            break;
        
        case CSDB_GROUP:
            dictionary = _groupDictionary;
            break;
            
        case CSDB_EVENT:
            dictionary = _eventDictionary;
            break;
            
        default:
            NSAssert(false, @"itemForType with invalid type");
            return nil;
    }
    
    CSDatabaseItem* item = [dictionary objectForKey:identifier];
    
    if (item == nil)
    {
        switch (type)
        {
            case CSDB_RELEASE:
                item = [[CSReleaseItem alloc] initWithReleaseId:identifier];
                break;
                
            case CSDB_SCENER:
                item = [[CSScenerItem alloc] initWithScenerId:identifier];
                break;
                
            case CSDB_GROUP:
                item = [[CSGroupItem alloc] initWithGroupId:identifier];
                break;

            case CSDB_EVENT:
                item = [[CSEventItem alloc] initWithEventId:identifier];
                break;

            default:
                NSAssert(false, @"itemForType with invalid type");
                return nil;
        }

        [dictionary setObject:item forKey:identifier];
    }

    return item;
}


// ----------------------------------------------------------------------------
- (void) writeTimeStampFile:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    NSString* path = [self cachePathForItem:item];

    NSTimeInterval timestamp = [NSDate timeIntervalSinceReferenceDate];
    NSData* timestampData = [NSData dataWithBytes:&timestamp length:sizeof(NSTimeInterval)];
    NSString* timeStampPath = [path stringByAppendingPathComponent:@"timestamp"];
    [timestampData writeToFile:timeStampPath atomically:YES];
}


// ----------------------------------------------------------------------------
- (NSTimeInterval) readTimeStamp:(NSString*)identifier ofType:(CSDatabaseItemType)type
// ----------------------------------------------------------------------------
{
    NSString* path = [self cachePathForId:identifier ofType:type];
    NSString* timeStampPath = [path stringByAppendingPathComponent:@"timestamp"];

    NSData* timestampData = [NSData dataWithContentsOfFile:timeStampPath];
    NSTimeInterval timestamp;
    [timestampData getBytes:&timestamp length:sizeof(NSTimeInterval)];
    return timestamp;
}


// ----------------------------------------------------------------------------
- (void) writeItem:(CSDatabaseItem*)item withData:(NSData*)data
// ----------------------------------------------------------------------------
{
    NSString* path = [self cachePathForItem:item];
    
    NSString* releaseDataPath = [path stringByAppendingPathComponent:@"data.xml"];
    [data writeToFile:releaseDataPath atomically:YES];

    [self writeTimeStampFile:item];
}


// ----------------------------------------------------------------------------
- (BOOL) readItem:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    // check if item has a cached directory with a data.xml file
    // read xml data and init item with it

    NSString* path = [self cachePathForItem:item];
    NSString* dataPath = [path stringByAppendingPathComponent:@"data.xml"];
    
    BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:dataPath isDirectory:NULL];
    if (exists)
    {
        NSData* data = [NSData dataWithContentsOfFile:dataPath];
        
        [item setFromXMLData:data];
        item.cached_depth = 2;
        return YES;
    }
    
    return NO;
}


// ----------------------------------------------------------------------------
- (void) writeDownloadFiles:(NSArray*)downloadFiles forReleaseId:(NSString*)releaseId
// ----------------------------------------------------------------------------
{
    NSString* path = [self cachePathForId:releaseId ofType:CSDB_RELEASE];

    NSString* downloadFilesString = @"";
    BOOL first_file = YES;
    for (NSString* downloadFile in downloadFiles)
    {
        NSString* relativePath = [downloadFile stringByReplacingOccurrencesOfString:path withString:@""];
        relativePath = [relativePath substringFromIndex:1];
        
        if (first_file)
            first_file = NO;
        else
            downloadFilesString = [downloadFilesString stringByAppendingString:@"\n"];
        
        downloadFilesString = [downloadFilesString stringByAppendingFormat:@"%@", relativePath];
    }
    
    NSString* downloadFilePath = [path stringByAppendingPathComponent:@"files.txt"];
    [downloadFilesString writeToFile:downloadFilePath atomically:YES encoding:NSUTF8StringEncoding error:NULL];
}


// ----------------------------------------------------------------------------
- (NSArray*) readDownloadFilesForReleaseId:(NSString*)releaseId
// ----------------------------------------------------------------------------
{
    NSString* path = [self cachePathForId:releaseId ofType:CSDB_RELEASE];
    NSString* downloadFilePath = [path stringByAppendingPathComponent:@"files.txt"];

    NSString* downloadFilesString = [NSString stringWithContentsOfFile:downloadFilePath encoding:NSUTF8StringEncoding error:NULL];
    if (downloadFilesString == nil)
        return nil;
    
    NSArray* downloadFiles = [downloadFilesString componentsSeparatedByString:@"\n"];
    return downloadFiles;
}


// ----------------------------------------------------------------------------
- (NSMutableArray*) favorites
// ----------------------------------------------------------------------------
{
    return _favorites;
}


// ----------------------------------------------------------------------------
- (void) writeFavorites
// ----------------------------------------------------------------------------
{
    [_favorites writeToFile:_favoritesPath atomically:YES];
}


// ----------------------------------------------------------------------------
- (void) readFavorites
// ----------------------------------------------------------------------------
{
    _favorites = [NSMutableArray arrayWithContentsOfFile:_favoritesPath];
    if (_favorites == nil)
    {
        _favorites = [NSMutableArray array];
        [self writeFavorites];
    }
}


// ----------------------------------------------------------------------------
- (void) addFavorite:(NSString*)releaseId
// ----------------------------------------------------------------------------
{
    if (![_favorites containsObject:releaseId])
    {
        [_favorites addObject:releaseId];
        [self writeFavorites];
    }
}


// ----------------------------------------------------------------------------
- (void) removeFavorite:(NSString*)releaseId
// ----------------------------------------------------------------------------
{
    if ([_favorites containsObject:releaseId])
    {
        [_favorites removeObject:releaseId];
        [self writeFavorites];
    }
}


// ----------------------------------------------------------------------------
- (NSString*) cachePathForItem:(CSDatabaseItem*)item
// ----------------------------------------------------------------------------
{
    return [self cachePathForId:item.identifier ofType:item.type];
}


// ----------------------------------------------------------------------------
- (NSString*) cachePathForId:(NSString*)identifier ofType:(CSDatabaseItemType)type
// ----------------------------------------------------------------------------
{
    switch (type)
    {
        case CSDB_RELEASE:
            return [self releasePathForReleaseId:identifier];
            break;
            
        case CSDB_SCENER:
            return [self scenerPathForScenerId:identifier];
            break;
        default:
            NSAssert(false, @"cachePathForId with invalid type");
            return nil;
    }
}


// ----------------------------------------------------------------------------
- (NSString*) releasePathForReleaseId:(NSString*)releaseId
// ----------------------------------------------------------------------------
{
    NSString* path = [_releasesPath stringByAppendingPathComponent:releaseId];

    BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:NULL];
    if (!exists)
        [[NSFileManager defaultManager] createDirectoryAtPath:path withIntermediateDirectories:YES attributes:nil error:NULL];
    
    return path;
}



// ----------------------------------------------------------------------------
- (NSString*) scenerPathForScenerId:(NSString*)scenerId
// ----------------------------------------------------------------------------
{
    NSString* path = [_scenersPath stringByAppendingPathComponent:scenerId];
    
    BOOL exists = [[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:NULL];
    if (!exists)
        [[NSFileManager defaultManager] createDirectoryAtPath:path withIntermediateDirectories:YES attributes:nil error:NULL];
    
    return path;
}


// ----------------------------------------------------------------------------
- (void) showNetworkError
// ----------------------------------------------------------------------------
{
    UIAlertView* alertView = [[UIAlertView alloc] initWithTitle:@"ERROR" message:@"Can't connect to CSDb server, check network settings or try again later." delegate:self cancelButtonTitle:nil otherButtonTitles:@"OK", nil];
    [alertView show];
}



static NSString* sMonthNames[13] =
{
    @"Unknown",
    @"January",
    @"February",
    @"March",
    @"April",
    @"May",
    @"June",
    @"July",
    @"August",
    @"September",
    @"October",
    @"November",
    @"December"
};

static NSString* sShortMonthNames[13] =
{
    @" ",
    @"Jan",
    @"Feb",
    @"Mar",
    @"Apr",
    @"May",
    @"Jun",
    @"Jul",
    @"Aug",
    @"Sep",
    @"Oct",
    @"Nov",
    @"Dec"
};


// ----------------------------------------------------------------------------
- (NSString*) stringForDate:(NSDateComponents*)date short:(BOOL)shortFormat
// ----------------------------------------------------------------------------
{
    NSInteger day = [date day];
    NSInteger month = [date month];
    NSInteger year = [date year];
    if (month > 12)
        month = 0;
    
    if (shortFormat)
    {
        if (day != 0 && month != 0 && year != 0)
            return [NSString stringWithFormat:@"%ld %@ %ld", (long)day, sShortMonthNames[month], (long)year];
        else if (month != 0 && year != 0)
            return [NSString stringWithFormat:@"%@ %ld", sShortMonthNames[month], (long)year];
        else if (month != 0 && year != 0)
            return [NSString stringWithFormat:@"%ld", (long)year];
    }
    else
    {
        if (day != 0 && month != 0 && year != 0)
            return [NSString stringWithFormat:@"%ld %@ %ld", (long)day, sMonthNames[month], (long)year];
        else if (month != 0 && year != 0)
            return [NSString stringWithFormat:@"%@ %ld", sMonthNames[month], (long)year];
        else if (month != 0 && year != 0)
            return [NSString stringWithFormat:@"%ld", (long)year];
    }
    
    return nil;
}



@end
